#
# Finite State Class
# Gbor SEBESTYN <segabor@elte.hu>
#

class FState
    # constants
    FS_START    = 0
    FS_RUN      = 1
    FS_END      = -1
    FS_SSET     = [ FS_START, FS_RUN, FS_END ]
    FS_TSET     = [ FS_END ]

    # ivars
    attr        :states    # set holding all state symbols
    attr        :edges     # hash table containing directional edges
    attr        :s         # start symbol
    attr        :t         # termination symbols
    attr        :state     # current state
    attr        :last       # last state
    attr_writer :delegate   # state delegate
    attr_writer :repeats   # allows repeated events?

    ## Factory
    def FState.default
        FState.new(FS_SSET, nil, FS_TSET, nil)
    end

    ## Default Constructor
    def initialize(state_set, start, term_set, state_edges)
        @delegate = self

        @states = state_set
        @t      = ( term_set.nil? ? [ @states[-1] ] : term_set )
        @s      = ( start.nil? ? @states[0] : start)

        @last   = nil
        @state  = @s

        if state_edges.nil?
            # edges are empty create a default one
            @edges = Hash.new
            @states[0...-1].each_index do |i|
                @edges[@states[i]] = @states[i+1]
            end
        else
            @edges = state_edges
        end
        
        @repeats = false
    end

    ## Set Next State
    def setState(ns)
        # leave if ns is nil or me terminated
        return if ns.nil? or self.terminated?
        # is state valid at all?
        unless @states.include? ns
            raise "#{self.id}.setState(#{ns}) invalid state!"
        end

        if ns == @state
            if @repeats == true
                # call state repeats
                meth = "rep_" + @state.to_s
                @delegate.method(meth).call if @delegate.methods.include? meth
            end
        else
            # let's see the graph edges
            nss = nextStates(@state)
            raise "#{self}.setState(#{ns}) nil edges or no edge starting with #{@state}!" unless nss.include? ns
            # call state to state delegate method
            meth = "allow_" + @state.to_s + "_to_" + @ns.to_s
            if @delegate.methods.include? meth
                b = @delegate.method(meth).call
                return unless b == true
            end
    
            # call exit_old_state delegate method
            meth = "exit_" + @state.to_s
            @delegate.method(meth).call if @delegate.methods.include? meth
            
            # set new state
            @last       = @state
            @state      = ns
    
            # call state to state changed delegate method
            meth = "move_" + @last.to_s + "_to_" + @state.to_s
            @delegate.method(meth).call if @delegate.methods.include? meth
    
            # call enter_new_state method
            meth = "enter_" + @state.to_s
            @delegate.method(meth).call if @delegate.methods.include? meth
        end
    end
    
    # step back to last state
    def back
        setState(@last)
    end
    
    def nextStates(st)
        raise "#{self}.nextStates(#{st}) invalid state!" unless @states.include? st

        # let's see the graph edges
        if not @edges.nil? and @edges.has_key? @state
            val = @edges[st]
        else
            raise "#{self}.nextStates(#{st}) nil edges or no edge starting with #{@state}!"
        end

        if val.is_a? Array
            val
        else
            [ val ]
        end
    end

    def start?
        @state == @s
    end

    def terminated?
        @t.include? state
    end
    
    def to_s
        "FState[#{id}]:#{state}"
    end

    def reset
        setState(@s)
    end
end
